@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
+ * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -40,17 +40,16 @@ int main(int argc, char *argv[])
 	int mode = 0;
 	int opt;
 	int option_index;
-	char *help_text;
 	char *c;
 	int rv;
 	int i;
 
 	int chunk = 0;
-	int size = -1;
+	long long size = -1;
 	int level = UnSet;
 	int layout = UnSet;
 	int raiddisks = 0;
-	int max_disks = MD_SB_DISKS;
+	int max_disks = MD_SB_DISKS; /* just a default */
 	int sparedisks = 0;
 	struct mddev_ident_s ident;
 	char *configfile = NULL;
@@ -63,6 +62,7 @@ int main(int argc, char *argv[])
 	int write_behind = 0;
 	int bitmap_fd = -1;
 	char *bitmap_file = NULL;
+	char *backup_file = NULL;
 	int bitmap_chunk = UnSet;
 	int SparcAdjust = 0;
 	mddev_dev_t devlist = NULL;
@@ -75,15 +75,21 @@ int main(int argc, char *argv[])
 	int force = 0;
 	int test = 0;
 	int assume_clean = 0;
-	int autof = 0; /* -2 means create device based on name:
-			*    if it ends mdN, then non-partitioned array N
-			*    if it ends dN, then partitions array N
-			* -1 means create non-partitioned, choose N
-			*  1 or more to create partitioned
-			* If -1 or 1 and name is a 'standard' name, then
-			* insist on a match of type and number.
-			*/
+	char *symlinks = NULL;
+	/* autof indicates whether and how to create device node.
+	 * bottom 3 bits are style.  Rest (when shifted) are number of parts
+	 * 0  - unset
+	 * 1  - don't create (no)
+	 * 2  - if is_standard, then create (yes)
+	 * 3  - create as 'md' - reject is_standard mdp (md)
+	 * 4  - create as 'mdp' - reject is_standard md (mdp)
+	 * 5  - default to md if not is_standard (md in config file)
+	 * 6  - default to mdp if not is_standard (part, or mdp in config file)
+	 */
+	int autof = 0;
 
+	char *homehost = NULL;
+	char sys_hostname[256];
 	char *mailaddr = NULL;
 	char *program = NULL;
 	int delay = 0;
@@ -93,8 +99,12 @@ int main(int argc, char *argv[])
 	struct supertype *ss = NULL;
 	int writemostly = 0;
 	int re_add = 0;
+	char *shortopt = short_options;
+	int dosyslog = 0;
+	int auto_update_home = 0;
 
 	int copies;
+	int print_help = 0;
 
 	int mdfd = -1;
 
@@ -109,32 +119,23 @@ int main(int argc, char *argv[])
 	ident.autof = 0;
 	ident.st = NULL;
 	ident.bitmap_fd = -1;
+	ident.bitmap_file = NULL;
 	ident.name[0] = 0;
 
 	while ((option_index = -1) ,
 	       (opt=getopt_long(argc, argv,
-				short_options, long_options,
+				shortopt, long_options,
 				&option_index)) != -1) {
 		int newmode = mode;
 		/* firstly, some mode-independant options */
 		switch(opt) {
 		case 'h':
-			help_text = Help;
 			if (option_index > 0 && 
 			    strcmp(long_options[option_index].name, "help-options")==0)
-				help_text = OptionHelp;
+				print_help = 2;
 			else
-				switch (mode) {
-				case ASSEMBLE : help_text = Help_assemble; break;
-				case BUILD    : help_text = Help_build; break;
-				case CREATE   : help_text = Help_create; break;
-				case MANAGE   : help_text = Help_manage; break;
-				case MISC     : help_text = Help_misc; break;
-				case MONITOR  : help_text = Help_monitor; break;
-				case GROW     : help_text = Help_grow; break;
-				}
-			fputs(help_text,stderr);
-			exit(0);
+				print_help = 1;
+			continue;
 
 		case 'V':
 			fputs(Version, stderr);
@@ -150,6 +151,14 @@ int main(int argc, char *argv[])
 			if (mode == ASSEMBLE || mode == BUILD || mode == CREATE || mode == GROW)
 				break; /* b means bitmap */
 			brief = 1;
+			if (optarg) {
+				fprintf(stderr, Name ": -b cannot have any extra immediately after it, sorry.\n");
+				exit(2);
+			}
+			continue;
+
+		case HomeHost:
+			homehost = optarg;
 			continue;
 
 		case ':':
@@ -164,19 +173,24 @@ int main(int argc, char *argv[])
 
 		switch(opt) {
 		case '@': /* just incase they say --manage */
-			newmode = MANAGE; break;
+			newmode = MANAGE;
+			shortopt = short_bitmap_auto_options;
+			break;
 		case 'a':
 		case 'r':
 		case 'f':
-		case 6: /* re-add */
-			if (!mode) newmode = MANAGE; 
+		case ReAdd: /* re-add */
+			if (!mode) {
+				newmode = MANAGE;
+				shortopt = short_bitmap_auto_options;
+			}
 			break;
 
-		case 'A': newmode = ASSEMBLE; break;
-		case 'B': newmode = BUILD; break;
-		case 'C': newmode = CREATE; break;
+		case 'A': newmode = ASSEMBLE; shortopt = short_bitmap_auto_options; break;
+		case 'B': newmode = BUILD; shortopt = short_bitmap_auto_options; break;
+		case 'C': newmode = CREATE; shortopt = short_bitmap_auto_options; break;
 		case 'F': newmode = MONITOR;break;
-		case 'G': newmode = GROW; break;
+		case 'G': newmode = GROW; shortopt = short_bitmap_auto_options; break;
 
 		case '#':
 		case 'D':
@@ -225,6 +239,7 @@ int main(int argc, char *argv[])
 					dv->disposition = devmode;
 					dv->writemostly = writemostly;
 					dv->re_add = re_add;
+					dv->used = 0;
 					dv->next = NULL;
 					*devlistend = dv;
 					devlistend = &dv->next;
@@ -302,6 +317,9 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+		case O(ASSEMBLE,AutoHomeHost):
+			auto_update_home = 1;
+			continue;
 		case O(CREATE,'e'):
 		case O(ASSEMBLE,'e'):
 		case O(MISC,'e'): /* set metadata (superblock) information */
@@ -336,7 +354,7 @@ int main(int argc, char *argv[])
 			if (strcmp(optarg, "max")==0)
 				size = 0;
 			else {
-				size = strtol(optarg, &c, 10);
+				size = strtoll(optarg, &c, 10);
 				if (!optarg[0] || *c || size < 4) {
 					fprintf(stderr, Name ": invalid size: %s\n",
 						optarg);
@@ -400,16 +418,18 @@ int main(int argc, char *argv[])
 				break;
 
 			case 10:
-				/* 'f' or 'n' followed by a number <= raid_disks */
-				if ((optarg[0] !=  'n' && optarg[0] != 'f') ||
+				/* 'f', 'o' or 'n' followed by a number <= raid_disks */
+				if ((optarg[0] !=  'n' && optarg[0] != 'f' && optarg[0] != 'o') ||
 				    (copies = strtoul(optarg+1, &cp, 10)) < 1 ||
 				    copies > 200 ||
 				    *cp) {
-					fprintf(stderr, Name ": layout for raid10 must be 'nNN' or 'fNN' where NN is a number, not %s\n", optarg);
+					fprintf(stderr, Name ": layout for raid10 must be 'nNN', 'oNN' or 'fNN' where NN is a number, not %s\n", optarg);
 					exit(2);
 				}
 				if (optarg[0] == 'n')
 					layout = 256 + copies;
+				else if (optarg[0] == 'o')
+					layout = 0x10000 + (copies<<8) + 1;
 				else
 					layout = 1 + (copies<<8);
 				break;
@@ -433,8 +453,8 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
-		case O(CREATE,3):
-		case O(BUILD,3): /* assume clean */
+		case O(CREATE,AssumeClean):
+		case O(BUILD,AssumeClean): /* assume clean */
 			assume_clean = 1;
 			continue;
 
@@ -477,39 +497,13 @@ int main(int argc, char *argv[])
 		case O(CREATE,'a'):
 		case O(BUILD,'a'):
 		case O(ASSEMBLE,'a'): /* auto-creation of device node */
-			if (optarg == NULL)
-				autof = -2;
-			else if (strcasecmp(optarg,"no")==0)
-				autof = 0;
-			else if (strcasecmp(optarg,"yes")==0)
-				autof = -2;
-			else if (strcasecmp(optarg,"md")==0)
-				autof = -1;
-			else {
-				/* There might be digits, and maybe a hypen, at the end */
-				char *e = optarg + strlen(optarg);
-				int num = 4;
-				int len;
-				while (e > optarg && isdigit(e[-1]))
-					e--;
-				if (*e) {
-					num = atoi(e);
-					if (num <= 0) num = 1;
-				}
-				if (e > optarg && e[-1] == '-')
-					e--;
-				len = e - optarg;
-				if ((len == 3 && strncasecmp(optarg,"mdp",3)==0) ||
-				    (len == 1 && strncasecmp(optarg,"p",1)==0) ||
-				    (len >= 4 && strncasecmp(optarg,"part",4)==0))
-					autof = num;
-				else {
-					fprintf(stderr, Name ": --auto flag arg of \"%s\" unrecognised: use no,yes,md,mdp,part\n"
-						"        optionally followed by a number.\n",
-						optarg);
-					exit(2);
-				}
-			}
+			autof = parse_auto(optarg, "--auto flag", 0);
+			continue;
+
+		case O(CREATE,Symlinks):
+		case O(BUILD,Symlinks):
+		case O(ASSEMBLE,Symlinks): /* auto creation of symlinks in /dev to /dev/md */
+			symlinks = optarg;
 			continue;
 
 		case O(BUILD,'f'): /* force honouring '-n 1' */
@@ -582,6 +576,12 @@ int main(int argc, char *argv[])
 				continue;
 			if (strcmp(update, "resync")==0)
 				continue;
+			if (strcmp(update, "uuid")==0)
+				continue;
+			if (strcmp(update, "name")==0)
+				continue;
+			if (strcmp(update, "homehost")==0)
+				continue;
 			if (strcmp(update, "byteorder")==0) {
 				if (ss) {
 					fprintf(stderr, Name ": must not set metadata type with --update=byteorder.\n");
@@ -596,9 +596,19 @@ int main(int argc, char *argv[])
 
 				continue;
 			}
-			fprintf(stderr, Name ": '--update %s' invalid.  Only 'sparc2.2', 'super-minor', 'resync' or 'summaries' supported\n",update);
+			if (strcmp(update,"?") == 0 || strcmp(update, "help") == 0)
+				fprintf(stderr, Name ": ");
+			else
+				fprintf(stderr, Name ": '--update=%s' is invalid.  ", update);
+			fprintf(stderr, "Valid --update options are:\n"
+				"     'sparc2.2', 'super-minor', 'uuid', 'name', 'resync',\n"
+				"     'summaries', 'homehost', 'byteorder'.\n");
 			exit(2);
 
+		case O(ASSEMBLE,NoDegraded): /* --no-degraded */
+			runstop = -1; /* --stop isn't allowed for --assemble, so we overload slightly */
+			continue;
+
 		case O(ASSEMBLE,'c'): /* config file */
 		case O(MISC, 'c'):
 		case O(MONITOR,'c'):
@@ -608,6 +618,7 @@ int main(int argc, char *argv[])
 				exit(2);
 			}
 			configfile = optarg;
+			set_conffile(configfile);
 			/* FIXME possibly check that config file exists.  Even parse it */
 			continue;
 		case O(ASSEMBLE,'s'): /* scan */
@@ -664,6 +675,10 @@ int main(int argc, char *argv[])
 		case O(MONITOR,'t'): /* test */
 			test = 1;
 			continue;
+		case O(MONITOR,'y'): /* log messages to syslog */
+			openlog("mdadm", 0, SYSLOG_FACILITY);
+			dosyslog = 1;
+			continue;
 
 			/* now the general management options.  Some are applicable
 			 * to other modes. None have arguments.
@@ -673,7 +688,7 @@ int main(int argc, char *argv[])
 			devmode = 'a';
 			re_add = 0;
 			continue;
-		case O(MANAGE,6):
+		case O(MANAGE,ReAdd):
 			devmode = 'a';
 			re_add = 1;
 			continue;
@@ -737,7 +752,7 @@ int main(int argc, char *argv[])
 			test = 1;
 			continue;
 
-		case O(MISC, 22):
+		case O(MISC, Sparc22):
 			if (devmode != 'E') {
 				fprintf(stderr, Name ": --sparc2.2 only allowed with --examine\n");
 				exit(2);
@@ -762,15 +777,34 @@ int main(int argc, char *argv[])
 			ident.bitmap_fd = bitmap_fd; /* for Assemble */
 			continue;
 
+		case O(ASSEMBLE, BackupFile):
+		case O(GROW, BackupFile):
+			/* Specify a file into which grow might place a backup,
+			 * or from which assemble might recover a backup
+			 */
+			if (backup_file) {
+				fprintf(stderr, Name ": backup file already specified, rejecting %s\n", optarg);
+				exit(2);
+			}
+			backup_file = optarg;
+			continue;
+
 		case O(GROW,'b'):
 		case O(BUILD,'b'):
 		case O(CREATE,'b'): /* here we create the bitmap */
-			bitmap_file = optarg;
-			continue;
+			if (strcmp(optarg, "internal")== 0 ||
+			    strcmp(optarg, "none")== 0 ||
+			    strchr(optarg, '/') != NULL) {
+				bitmap_file = optarg;
+				continue;
+			}
+			/* probable typo */
+			fprintf(stderr, Name ": bitmap file must contain a '/', or be 'internal', or 'none'\n");
+			exit(2);
 
-		case O(GROW,4):
-		case O(BUILD,4):
-		case O(CREATE,4): /* bitmap chunksize */
+		case O(GROW,BitmapChunk):
+		case O(BUILD,BitmapChunk):
+		case O(CREATE,BitmapChunk): /* bitmap chunksize */
 			bitmap_chunk = strtol(optarg, &c, 10);
 			if (!optarg[0] || *c || bitmap_chunk < 0 ||
 					bitmap_chunk & (bitmap_chunk - 1)) {
@@ -782,8 +816,8 @@ int main(int argc, char *argv[])
 			bitmap_chunk = bitmap_chunk ? bitmap_chunk * 1024 : 512;
 			continue;
 
-		case O(BUILD, 5):
-		case O(CREATE, 5): /* write-behind mode */
+		case O(BUILD, WriteBehind):
+		case O(CREATE, WriteBehind): /* write-behind mode */
 			write_behind = DEFAULT_MAX_WRITE_BEHIND;
 			if (optarg) {
 				write_behind = strtol(optarg, &c, 10);
@@ -798,12 +832,35 @@ int main(int argc, char *argv[])
 		/* We have now processed all the valid options. Anything else is
 		 * an error
 		 */
-		fprintf(stderr, Name ": option %c not valid in %s mode\n",
-			opt, map_num(modes, mode));
+		if (option_index > 0)
+			fprintf(stderr, Name ":option --%s not valid in %s mode\n",
+				long_options[option_index].name,
+				map_num(modes, mode));
+		else
+			fprintf(stderr, Name ": option -%c not valid in %s mode\n",
+				opt, map_num(modes, mode));
 		exit(2);
 
 	}
 
+	if (print_help) {
+		char *help_text = Help;
+		if (print_help == 2)
+			help_text = OptionHelp;
+		else
+			switch (mode) {
+			case ASSEMBLE : help_text = Help_assemble; break;
+			case BUILD    : help_text = Help_build; break;
+			case CREATE   : help_text = Help_create; break;
+			case MANAGE   : help_text = Help_manage; break;
+			case MISC     : help_text = Help_misc; break;
+			case MONITOR  : help_text = Help_monitor; break;
+			case GROW     : help_text = Help_grow; break;
+			}
+		fputs(help_text,stderr);
+		exit(0);
+	}
+
 	if (!mode && devs_found) {
 		mode = MISC;
 		devmode = 'Q';
@@ -814,6 +871,19 @@ int main(int argc, char *argv[])
 		fputs(Usage, stderr);
 		exit(2);
 	}
+
+	if (symlinks) {
+		struct createinfo *ci = conf_get_create_info();
+
+		if (strcasecmp(symlinks, "yes") == 0)
+			ci->symlinks = 1;
+		else if (strcasecmp(symlinks, "no") == 0)
+			ci->symlinks = 0;
+		else {
+			fprintf(stderr, Name ": option --symlinks must be 'no' or 'yes'\n");
+			exit(2);
+		}
+	}
 	/* Ok, got the option parsing out of the way
 	 * hopefully it's mostly right but there might be some stuff
 	 * missing
@@ -834,6 +904,8 @@ int main(int argc, char *argv[])
 			fprintf(stderr, Name ": --super-minor=dev is incompatible with --auto\n");	
 			exit(2);
 		}
+		if (mode == MANAGE || mode == GROW)
+			autof=1; /* Don't create */
 		mdfd = open_mddev(devlist->devname, autof);
 		if (mdfd < 0)
 			exit(1);
@@ -865,6 +937,15 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	if (homehost == NULL)
+		homehost = conf_get_homehost();
+	if (homehost && strcmp(homehost, "<system>")==0) {
+		if (gethostname(sys_hostname, sizeof(sys_hostname)) == 0) {
+			sys_hostname[sizeof(sys_hostname)-1] = 0;
+			homehost = sys_hostname;
+		}
+	}
+
 	rv = 0;
 	switch(mode) {
 	case MANAGE:
@@ -877,13 +958,13 @@ int main(int argc, char *argv[])
 		if (!rv && readonly < 0)
 			rv = Manage_ro(devlist->devname, mdfd, readonly);
 		if (!rv && runstop)
-			rv = Manage_runstop(devlist->devname, mdfd, runstop, 0);
+			rv = Manage_runstop(devlist->devname, mdfd, runstop, quiet);
 		break;
 	case ASSEMBLE:
 		if (devs_found == 1 && ident.uuid_set == 0 &&
 		    ident.super_minor == UnSet && ident.name[0] == 0 && !scan ) {
 			/* Only a device has been given, so get details from config file */
-			mddev_ident_t array_ident = conf_get_ident(configfile, devlist->devname);
+			mddev_ident_t array_ident = conf_get_ident(devlist->devname);
 			if (array_ident == NULL) {
 				fprintf(stderr, Name ": %s not identified in config file.\n",
 					devlist->devname);
@@ -894,23 +975,27 @@ int main(int argc, char *argv[])
 				if (mdfd < 0)
 					rv |= 1;
 				else {
-					rv |= Assemble(ss, devlist->devname, mdfd, array_ident, configfile,
-						       NULL,
-						       readonly, runstop, update, verbose-quiet, force);
+					rv |= Assemble(ss, devlist->devname, mdfd, array_ident,
+						       NULL, backup_file,
+						       readonly, runstop, update, homehost, verbose-quiet, force);
 					close(mdfd);
 				}
 			}
 		} else if (!scan)
-			rv = Assemble(ss, devlist->devname, mdfd, &ident, configfile,
-				      devlist->next,
-				      readonly, runstop, update, verbose-quiet, force);
+			rv = Assemble(ss, devlist->devname, mdfd, &ident,
+				      devlist->next, backup_file,
+				      readonly, runstop, update, homehost, verbose-quiet, force);
 		else if (devs_found>0) {
 			if (update && devs_found > 1) {
 				fprintf(stderr, Name ": can only update a single array at a time\n");
 				exit(1);
 			}
+			if (backup_file && devs_found > 1) {
+				fprintf(stderr, Name ": can only assemble a single array when providing a backup file.\n");
+				exit(1);
+			}
 			for (dv = devlist ; dv ; dv=dv->next) {
-				mddev_ident_t array_ident = conf_get_ident(configfile, dv->devname);
+				mddev_ident_t array_ident = conf_get_ident(dv->devname);
 				if (array_ident == NULL) {
 					fprintf(stderr, Name ": %s not identified in config file.\n",
 						dv->devname);
@@ -923,39 +1008,102 @@ int main(int argc, char *argv[])
 					rv |= 1;
 					continue;
 				}
-				rv |= Assemble(ss, dv->devname, mdfd, array_ident, configfile,
-					       NULL,
-					       readonly, runstop, update, verbose-quiet, force);
+				rv |= Assemble(ss, dv->devname, mdfd, array_ident,
+					       NULL, backup_file,
+					       readonly, runstop, update, homehost, verbose-quiet, force);
 				close(mdfd);
 			}
 		} else {
-			mddev_ident_t array_list =  conf_get_ident(configfile, NULL);
-			if (!array_list) {
+			mddev_ident_t array_list =  conf_get_ident(NULL);
+			mddev_dev_t devlist = conf_get_devs();
+			int cnt = 0;
+			if (devlist == NULL) {
+				fprintf(stderr, Name ": No devices listed in conf file were found.\n");
+				exit(1);
+			}
+			if (update) {
+				fprintf(stderr, Name ": --update not meaningful with a --scan assembly.\n");
+				exit(1);
+			}
+			if (backup_file) {
+				fprintf(stderr, Name ": --backup_file not meaningful with a --scan assembly.\n");
+				exit(1);
+			}
+			for (; array_list; array_list = array_list->next) {
+				mdu_array_info_t array;
+				mdfd = open_mddev(array_list->devname,
+						  array_list->autof ? array_list->autof : autof);
+				if (mdfd < 0) {
+					rv |= 1;
+					continue;
+				}
+				if (ioctl(mdfd, GET_ARRAY_INFO, &array)>=0)
+					/* already assembled, skip */
+					cnt++;
+				else {
+					rv |= Assemble(ss, array_list->devname, mdfd,
+						       array_list,
+						       NULL, NULL,
+						       readonly, runstop, NULL, homehost, verbose-quiet, force);
+					if (rv == 0) cnt++;
+				}
+				close(mdfd);
+			}
+			if (homehost) {
+				/* Maybe we can auto-assemble something.
+				 * Repeatedly call Assemble in auto-assmble mode
+				 * until it fails
+				 */
+				int rv2;
+				int acnt;
+				ident.autof = autof;
+				do {
+					acnt = 0;
+					do {
+						rv2 = Assemble(ss, NULL, -1,
+							       &ident,
+							       NULL, NULL,
+							       readonly, runstop, NULL, homehost, verbose-quiet, force);
+						if (rv2==0) {
+							cnt++;
+							acnt++;
+						}
+						if (rv2 == 1)
+							/* found something so even though assembly failed  we
+							 * want to avoid auto-updates
+							 */
+							auto_update_home = 0;
+					} while (rv2!=2);
+					/* Incase there are stacked devices, we need to go around again */
+				} while (acnt);
+				if (cnt == 0 && auto_update_home && homehost) {
+					/* Nothing found, maybe we need to bootstrap homehost info */
+					do {
+						acnt = 0;
+						do {
+							rv2 = Assemble(ss, NULL, -1,
+								       &ident,
+								       NULL, NULL,
+								       readonly, runstop, "homehost", homehost, verbose-quiet, force);
+							if (rv2==0) {
+								cnt++;
+								acnt++;
+							}
+						} while (rv2!=2);
+						/* Incase there are stacked devices, we need to go around again */
+					} while (acnt);
+				}
+				if (cnt == 0 && rv == 0) {
+					fprintf(stderr, Name ": No arrays found in config file or automatically\n");
+					rv = 1;
+				}
+			} else if (cnt == 0 && rv == 0) {
 				fprintf(stderr, Name ": No arrays found in config file\n");
 				rv = 1;
-			} else
-				for (; array_list; array_list = array_list->next) {
-					mdu_array_info_t array;
-					mdfd = open_mddev(array_list->devname, 
-							  array_list->autof ? array_list->autof : autof);
-					if (mdfd < 0) {
-						rv |= 1;
-						continue;
-					}
-					if (ioctl(mdfd, GET_ARRAY_INFO, &array)>=0)
-						/* already assembled, skip */
-						;
-					else
-						rv |= Assemble(ss, array_list->devname, mdfd,
-							       array_list, configfile,
-							       NULL,
-							       readonly, runstop, NULL, verbose-quiet, force);
-					close(mdfd);
-				}
+			}
 		}
 		break;
 	case BUILD:
-		if (bitmap_chunk == UnSet) bitmap_chunk = DEFAULT_BITMAP_CHUNK;
 		if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
 		if (write_behind && !bitmap_file) {
 			fprintf(stderr, Name ": write-behind mode requires a bitmap.\n");
@@ -981,18 +1129,10 @@ int main(int argc, char *argv[])
 			rv = 1;
 			break;
 		}
-		if (ss == NULL) {
-			for(i=0; !ss && superlist[i]; i++) 
-				ss = superlist[i]->match_metadata_desc("default");
-		}
-		if (!ss) {
-			fprintf(stderr, Name ": internal error - no default metadata style\n");
-			exit(2);
-		}
 
 		rv = Create(ss, devlist->devname, mdfd, chunk, level, layout, size<0 ? 0 : size,
-			    raiddisks, sparedisks, ident.name,
-			    devs_found-1, devlist->next, runstop, verbose-quiet, force,
+			    raiddisks, sparedisks, ident.name, homehost,
+			    devs_found-1, devlist->next, runstop, verbose-quiet, force, assume_clean,
 			    bitmap_file, bitmap_chunk, write_behind, delay);
 		break;
 	case MISC:
@@ -1003,17 +1143,19 @@ int main(int argc, char *argv[])
 				exit(2);
 			}
 			if (devlist == NULL)
-				devlist = conf_get_devs(configfile);
+				devlist = conf_get_devs();
 			if (devlist == NULL) {
 				fprintf(stderr, Name ": No devices listed in %s\n", configfile?configfile:DefaultConfFile);
 				exit(1);
 			}
-			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief, scan, SparcAdjust, ss);
+			if (brief && verbose)
+				brief = 2;
+			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief, scan, SparcAdjust, ss, homehost);
 		} else {
 			if (devlist == NULL) {
 				if (devmode=='D' && scan) {
 					/* apply --detail to all devices in /proc/mdstat */
-					struct mdstat_ent *ms = mdstat_read(0);
+					struct mdstat_ent *ms = mdstat_read(0, 1);
 					struct mdstat_ent *e;
 					for (e=ms ; e ; e=e->next) {
 						char *name = get_md_name(e->devnum);
@@ -1023,7 +1165,7 @@ int main(int argc, char *argv[])
 								e->dev);
 							continue;
 						}
-						rv |= Detail(name, verbose>1?0:verbose+1, test);
+						rv |= Detail(name, verbose>1?0:verbose+1, test, homehost);
 						put_md_name(name);
 					}
 				} else	if (devmode == 'S' && scan) {
@@ -1034,7 +1176,7 @@ int main(int argc, char *argv[])
 					int progress=1, err;
 					int last = 0;
 					do {
-						struct mdstat_ent *ms = mdstat_read(0);
+						struct mdstat_ent *ms = mdstat_read(0, 0);
 						struct mdstat_ent *e;
 
 						if (!progress) last = 1;
@@ -1047,9 +1189,9 @@ int main(int argc, char *argv[])
 									e->dev);
 								continue;
 							}
-							mdfd = open_mddev(name, 0);
+							mdfd = open_mddev(name, 1);
 							if (mdfd >= 0) {
-								if (Manage_runstop(name, mdfd, -1, !last))
+								if (Manage_runstop(name, mdfd, -1, quiet?1:last?0:-1))
 									err = 1;
 								else
 									progress = 1;
@@ -1059,6 +1201,7 @@ int main(int argc, char *argv[])
 							put_md_name(name);
 						}
 					} while (!last && err);
+					if (err) rv |= 1;
 				} else {
 					fprintf(stderr, Name ": No devices given.\n");
 					exit(2);
@@ -1067,28 +1210,29 @@ int main(int argc, char *argv[])
 			for (dv=devlist ; dv; dv=dv->next) {
 				switch(dv->disposition) {
 				case 'D':
-					rv |= Detail(dv->devname, brief?1+verbose:0, test); continue;
+					rv |= Detail(dv->devname, brief?1+verbose:0, test, homehost); continue;
 				case 'K': /* Zero superblock */
-					rv |= Kill(dv->devname, force); continue;
+					rv |= Kill(dv->devname, force, quiet); continue;
 				case 'Q':
 					rv |= Query(dv->devname); continue;
 				case 'X':
 					rv |= ExamineBitmap(dv->devname, brief, ss); continue;
 				}
-				mdfd = open_mddev(dv->devname, 0);
+				mdfd = open_mddev(dv->devname, 1);
 				if (mdfd>=0) {
 					switch(dv->disposition) {
 					case 'R':
-						rv |= Manage_runstop(dv->devname, mdfd, 1, 0); break;
+						rv |= Manage_runstop(dv->devname, mdfd, 1, quiet); break;
 					case 'S':
-						rv |= Manage_runstop(dv->devname, mdfd, -1, 0); break;
+						rv |= Manage_runstop(dv->devname, mdfd, -1, quiet); break;
 					case 'o':
 						rv |= Manage_ro(dv->devname, mdfd, 1); break;
 					case 'w':
 						rv |= Manage_ro(dv->devname, mdfd, -1); break;
 					}
 					close(mdfd);
-				}
+				} else
+					rv |= 1;
 			}
 		}
 		break;
@@ -1104,7 +1248,8 @@ int main(int argc, char *argv[])
 			break;
 		}
 		rv= Monitor(devlist, mailaddr, program,
-			    delay?delay:60, daemonise, scan, oneshot, configfile, test, pidfile);
+			    delay?delay:60, daemonise, scan, oneshot,
+			    dosyslog, test, pidfile);
 		break;
 
 	case GROW:
@@ -1128,11 +1273,12 @@ int main(int argc, char *argv[])
 		} else if (layout != UnSet)
 			rv = Manage_reconfig(devlist->devname, mdfd, layout);
 		else if (size >= 0 || raiddisks)
-			rv = Manage_resize(devlist->devname, mdfd, size, raiddisks);
+			rv = Grow_reshape(devlist->devname, mdfd, quiet, backup_file,
+					  size, level, layout, chunk, raiddisks);
 		else if (bitmap_file) {
 			if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
 			rv = Grow_addbitmap(devlist->devname, mdfd, bitmap_file,
-					    bitmap_chunk, delay, write_behind);
+					    bitmap_chunk, delay, write_behind, force);
 		} else
 			fprintf(stderr, Name ": no changes to --grow\n");
 		break;
